#define MODULE
#define __KERNEL__

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <asm/uaccess.h>
#include <asm/vr41xx.h>
#include <asm/io.h>
#include <linux/delay.h>

#define AD_MAJOR	18
int ad_major = AD_MAJOR;

#define AD_CHANNEL(minor) ((minor) & 3)
#define AD_BINARY(minor)  ((minor) & 4)

void regbset(volatile unsigned short *port,unsigned mask,unsigned bits)
{
  unsigned short data0,data;
  data0 = readw(port);
  data = (data0 & ~mask) | (bits & mask);
  writew(data,port);
}

int ad_open(struct inode *inode, struct file *filp)
{
  int minor = MINOR(inode->i_rdev);
  regbset(VR41XX_CMUCLKMSK,
	  VR41XX_CMUCLKMSK_MSKPIUPCLK
	  |VR41XX_CMUCLKMSK_MSKADUPCLK 
	  |VR41XX_CMUCLKMSK_MSKADU18M
	  ,0xff);
  regbset(VR41XX_PIUCNTREG, 1, 1);   /* RST=1 */
  regbset(VR41XX_PIUCNTREG, 2, 2);   /* PWR=1 */
  MOD_INC_USE_COUNT;
  return 0;
}

int ad_release(struct inode *inode, struct file *filp)
{
  MOD_DEC_USE_COUNT;
  return 0;
}

ssize_t ad_read(struct file *filp, char *buf, size_t count,loff_t *f_pos)
{
  int minor = MINOR(filp->f_dentry->d_inode->i_rdev);
  int ch = AD_CHANNEL(minor);
  unsigned short data;
  size_t count1=0;
  int i;
  
  regbset(VR41XX_PIUINTREG, 0x7f, 0x7f);
  regbset(VR41XX_PIUCNTREG, 0x1c, 0x08);    /* MODE=01, SEQEN=0 */
  regbset(VR41XX_PIUCMDREG, 0x100f, ch | 0x1004);
  regbset(VR41XX_PIUCNTREG, 4, 4);          /* SEQEN=1 */
  i=0;
  do {
    data = readw(VR41XX_PIUAB0REG);
  } while(!(data & 0x8000) && i++ < 10000);
  
  regbset(VR41XX_PIUCNTREG, 4, 0);          /* SEQEN=0 */
  if(AD_BINARY(minor)){
    count1 = sizeof(unsigned short);
    if(count < count1){
      printk(KERN_INFO "ad_read: buffer size must be greater than %d\n",
	     count1-1);
      return -EINVAL;
    }
    if(put_user(data,(unsigned short *)buf)) return -EFAULT;
  }
  else {
    char sbuf[10];
    sprintf(sbuf,"%3d\n",data & 0x7fff);
    count1 = strlen(sbuf);
    if(count < count1){
      printk(KERN_INFO "ad_read: buffer size must be greater than %d\n",count1-1);
      return -EINVAL;
    }
    if(copy_to_user(buf, sbuf, count1)) return -EFAULT;
  }
  *f_pos += count1;
  return count1;
}

struct file_operations ad_fops = {
  read:    ad_read,  
  open:    ad_open,
  release: ad_release,
};

int init_module(void)
{
  int r;
  r = register_chrdev(ad_major, "ad", &ad_fops);
  if(r < 0){
    printk(KERN_WARNING "ad: can't get major %d\n", ad_major);
    return r;
  }
  if(ad_major == 0) ad_major = r;
  printk(KERN_INFO "ad installed\n");
  return 0;
}

void cleanup_module(void)
{
  int r;
  r = unregister_chrdev(ad_major, "ad");
  if(r < 0){
    printk(KERN_ALERT "ad: unregister major %d failed\n", ad_major);
    return;
  }
  printk(KERN_INFO "ad uninstalled\n");
}
